  (* ********************************************************  *)
  (*                                                           *)
  (* FRISTENSKY SEQUENCE ANALYSIS PACKAGE SUBROUTINE MODULES   *)
  (*             VERSION  September 26, 2023 Free Pascal            *)
  (*             Brian Fristensky                              *)
  (*             Dept. of Plant Science                        *)
  (*             University of Manitoba                        *)
  (*             Winnipeg, MB R3T 2N2 CANADA                   *)
  (*                                                           *)
  (* Copyright (c) 2022  by Brian Fristensky                   *)
  (* !!! in comment indicates feature which may need change.   *)
  (*  *******************************************************  *)
 
  (* Main procedure tests most (but not all) of the code.
     Sept. 26, 2023 
         - Added test for topology of nucleotide sequence
         - fixed bug in READSEQ that failed to detect circular topology
           in GenBank files.
     Oct. 2022 - code adapted to Free Pascal
        One major change is that most integers have been 
        changed to longint.
     Oct. 2022 - code added to terminal input routines to
        avoid infinite loops by terminating after several tries.
  *)

  program CSAPMOD(input, output (*,INFILE*));
(*!!! Some Pascals require file parameters in program heading *)

(* tell fpc compiler that this is a console application, to allow ctrl-C to
   terminate a program *)
uses 
   Sysutils, Crt;
 
  const MAXSEQ  = 250000;
(* BEGIN MODULE VERSION *)
        VERSION= 'XYLEMMODS     Version  May 24, 2023';
(* END MODULE VERSION *)
 
(* BEGIN MODULE REALLIMITS *)
(*!!!*)MAXREAL = 1.7E38;
(* END MODULE REALLIMITS *)
 
(* BEGIN MODULE INTLIMITS *)
(*!!!*)  MAXINT =  2147483647;
(*!!! *) MININT = -2147483647;
(* END MODULE INTLIMITS *)
 
(* BEGIN MODULE STARTARGNUM *)
	STARTARGNUM=1;    (* Free Pascal: ARG(1) is 1st command line argument*)
      (*STARTARGNUM=2;       ATT Pascal: ARG(2) is 1st command line argument*)
(* END MODULE STARTARGNUM *)

        MAXWORD = 25;
        MAXLINE = 70;
        MAXRANGE=  0;

  type NUCLEOTIDE  = (CANT,A,C,R,D,V,M,K,B,H,Y,G,T,W,S,N,Z,WONT);
       AMINOACID   = (GLY,ALA,VALINE,LEU,ILE,MET,PHE,PRO,SER,THR,CYS,ASN,GLN,
                      TYR,TRP,ASP,GLU,HIS,LYS,ARG,ASX,GLX,TERM,UNKX,UNKY);
       SEQUENCE    = array[1..MAXSEQ] of NUCLEOTIDE;
       PROTEIN     = array[1..10000] of AMINOACID;
       CHSET       = set of 'A'..'Z';
       LETTERS = packed array[1..10] of char;
       VECTOR  = array[1..MAXLINE] of integer;
 
(* BEGIN MODULE TYPE.WORD *)
       (*   <word>::= <non-blank char>[<non-blank char>] *)
       WORD    = record
                 LEN:integer;
                 STR:array[1..MAXWORD] of char
                 end;
(* END MODULE TYPE.WORD         VERSION= 'CSAPMODS     Version  4/13/88'; *)
 
(* BEGIN MODULE TYPE.LINE *)
       CHARARRAY = array[1..MAXLINE] of char;
       LINE = record
                INPSTR:string;
                STR:CHARARRAY;
                LEN:integer
                end;
(* END MODULE TYPE.LINE         VERSION= 'SUNMODS     Version  3/21/90'; *)
 
(* BEGIN MODULE TYPE.FRAG *)
       (* LIST OF RESTRICTION SITES FOUND *)
       FRAG = ^FRAGMENT;
       FRAGMENT = record
         START,FINISH,SIZE:longint;
         PREV,NEXT:FRAG
         end;
       FRAGSFOUND = record
         LNUM:integer;
         HEAD,TAIL:FRAG
         end;
(* END MODULE TYPE.FRAG *)
 
  var INFILE,OUTFILE: text;
      CH:char;
      WO : WORD;
      LI : LINE;
      NUMWORDS, I : integer; (* number of words in an input line *)
      RE:real;
      SEQ         : SEQUENCE;
      PROT        : PROTEIN;
      SEQLEN,INT  : longint;
      CIRCULAR    : boolean;
      FREEFRAG:FRAG;
      ORDER: array[1..10] of FRAG;
      RESULT: integer;
 
 
(********************* ACTUAL PROCEDURES ***************)
   (**********************************************************)
   (*  LINE I/O Procedures.                                  *)
   (**********************************************************)

(* BEGIN MODULE INPLINE *) 
    (*  Read a line from the console.    *)
    procedure INPLINE(var W:LINE);
      var I : integer;
      begin
        with W do begin
            readln(input,INPSTR);
            I := 1;
            LEN := Length(INPSTR);
            while I <= LEN do begin
                STR[I] := INPSTR[I];
                I := I + 1
            end
        end
      end; (* INPLINE *)
(* END MODULE INPLINE *)

(* BEGIN MODULE READLINE *)
  (* Read a line from a file, omitting trailing blanks *)  
  procedure READLINE(var F:text; var L:LINE);
    var LASTNONBLANK,I:integer;
        CH:char;
    begin
      with L do begin
        LEN:=0; LASTNONBLANK:=0;
        while not eoln(F) do begin
          read(F,CH);
          if LEN < MAXLINE then begin 
            LEN:=LEN+1; STR[LEN]:=CH;
            if CH <> ' ' then LASTNONBLANK:=LEN
            end
          end;
        if not eof(F) then readln(F);
        LEN:=LASTNONBLANK;
        for I:= LEN+1 to MAXLINE do STR[I]:=' '
        end; (* with L*) 
    end; (* READLINE *)
(* END MODULE READLINE *)

(* BEGIN MODULE LINEPOSN *)
   (*  return the leftmost position of CH in L 
     return -1 if not found. *)
   function LINEPOSN(var L:LINE; CH:char):integer;
     var I :integer;
         FOUND: integer;
     begin
       FOUND := -1;
       I := 1;
       with L  do begin
         while (I <= LEN) and (FOUND = -1) do begin
           if (STR[I] = CH) then FOUND := I;
           I := I + 1
           end (* while *)
       end; (* with *)   
       LINEPOSN := FOUND       
     end; (* LINEPOSN *)
(* END MODULE LINEPOSN VERSION= 'CSAPMODS     Version  April 5, 2023'; *)

(* BEGIN MODULE NEXTWORD *)
  (* Read the next word from a line. I is the next 
       position in the line that has not been read. *)  
  procedure NEXTWORD(var L:LINE; var W:WORD; var I:integer);
    var J:integer;
    begin
        with W do begin
            LEN := 0;
            while (L.STR[I] = ' ') and (I < L.LEN) do I := I + 1; (* skip leading blanks *)
            if I <= L.LEN then
                while (L.STR[I] <> ' ') and (I <= L.LEN) do begin
                    LEN:=LEN+1; STR[LEN]:= L.STR[I]; I := I + 1
                    end;
            for J:= LEN+1 to MAXWORD do STR[J]:= ' ' 
            end (* with W*) 
    end; (* NEXTWORD *)
(* END MODULE NEXTWORD *)

(* BEGIN MODULE NUMBER *)
  (* Extract an integer from a CHARARRAY (see TYPE.LINE), starting
     at the current position. *)
  function NUMBER(TARGET:CHARARRAY; var POS:integer):longint;
    var N:longint;
        ORDZERO:integer;
    begin
      ORDZERO:=ord('0');
      N:=0;
      (* evaluate characteristic *)
      while TARGET[POS] in ['0'..'9'] do begin
        N:= (N*10) + (ord(TARGET[POS])-ORDZERO);
        POS:=POS+1
        end; (* while *)
      NUMBER:=N
    end; (* NUMBER *)
(* END MODULE NUMBER *)

(* BEGIN MODULE WRITELINE *)
   (*  Write a line to a file using L char, left-justified.  *)
   procedure WRITELINE(var F:text;W:LINE;L:integer);
     var I :integer;
     begin
       with W do
         for I := 1 to L do
           if I <= LEN then write(F,STR[I])
           else write(F,' ')
     end; (* WRITELINE *)
(* END MODULE WRITELINE         VERSION= 'CSAPMODS     Version  4/13/88'; *)

(* BEGIN MODULE GETSUBSTR *)
   (*  return a LINE object containing substring of a LINE,
       from positions FIRST to LAST  *)
   function GETSUBSTR(W:LINE;FIRST,LAST:integer):LINE;
     var I,J :integer;
         L : LINE;
     begin
       L.LEN := 0;
       J := 1;
       with W do
         for I := FIRST to LAST do begin
             if (I <= LEN) then L.STR[J] := STR[I];
                 J := J +1;
                 L.LEN := L.LEN + 1
                 end; (* for *)
       for J := L.LEN + 1 to MAXLINE do L.STR[J] := ' ';
      GETSUBSTR := L
     end; (* GETSUBSTR *)
(* END MODULE GETSUBSTR         VERSION= 'CSAPMODS     Version  Apr. 27, 2023'; *)

(* BEGIN MODULE WRITESUBSTR *)
   (*  Write a  substring of line to a file, from positions FIRST to LAST  *)
   procedure WRITESUBSTR(var F:text;W:LINE;FIRST,LAST:integer);
     var I :integer;
     begin
       with W do
         for I := FIRST to LAST do
           if I <= LEN then write(F,STR[I])
           else write(F,' ')
     end; (* WRITESUBSTR *)
(* END MODULE WRITESUBSTR         VERSION= 'CSAPMODS     Version  Apr. 25, 2023'; *)

(* BEGIN MODULE FILEARGS *)
  (* This procedure overcomes one of the stupidest aspects of UNIX Pascal,
     namely the fact that filenames in the program statement are supposed to
     be actual UNIX filenames!  To overcome this, the 2-argument version of
     reset and rewrite must be used with string variables.  This module
     need only contain the reset and rewrite statements in any normal 
     implementation of Pascal. *)
  procedure FILEARGS(var F:text; FTYPE:char; var ARGNUM:integer);
    var  ARGUMENT : shortString;
    begin
      ARGUMENT := paramStr(ARGNUM);
      Assign(F,ARGUMENT);
      if FTYPE='I' then reset(F)
      else rewrite(F);
      ARGNUM:=ARGNUM+1
    end; (* FILEARGS *)
(* END MODULE FILEARGS *)

(* BEGIN MODULE GETFILE *)

(************************************************************)
(* Open  files, checking for I/O errors.                    *)
(* Assumes user has already been prompted for filename.     *)
(* Syntax: GETFILE(F,'I',<filename>);                       *)
(*              opens file F for input                      *)
(*         GETFILE(F,'O',<filename>);                       *)
(*              opens file F for output                     *)
(* !!! NOTE: this non-Standard procedure must be rewritten  *)
(* for each different version of Pascal.                    *)
(************************************************************)
procedure GETFILE(var F:text; FTYPE:char; var FILENAME:LINE);

  const MAXATTEMPTS = 5;

  var CH:char;
      OKAY:boolean;
      ATTEMPTS : integer;

  begin (* GETFILE - - - - - - - - - - - - - - - - - - - - - *)
      (* Read in filename.*)
      INPLINE(FILENAME);
      ATTEMPTS := 1;
      case FTYPE of
        (* Input file must exist before being opened.*)
        'I':begin (* input file *)
              while not(FileExists(FILENAME.INPSTR)) do begin
                writeln('***FILE NOT FOUND');
                writeln('Enter another filename:');
                (* Prevent an infinite loop. *)
                if ATTEMPTS > MAXATTEMPTS then begin
                    writeln('>>> Too many attempts. Terminating program.');
                    Halt
                    end
                else begin
                    INPLINE(FILENAME);
                    ATTEMPTS := ATTEMPTS + 1
                    end
                end; (* while *)
            Assign(F,FILENAME.INPSTR);
(*!!!*)     reset(F)
            end; (* I *)

        (* If output file already exists, ask user if he/she/they really want
           to overwrite it. *)
        'O':begin
              repeat
                OKAY:=true;
                if FileExists(FILENAME.INPSTR) then begin
                  repeat  
                    writeln('*** WARNING! File already exists. Overwrite?[Y|N]');
                    readln(CH);
                    if not ( CH in ['Y','y']) then begin
                        (* Prevent an infinite loop. *)
                        if ATTEMPTS > MAXATTEMPTS then begin
                             writeln('>>> Too many attempts. Terminating program.');
                            Halt
                            end
                        else begin
                            writeln('Enter another filename:');
                            INPLINE(FILENAME);
                            ATTEMPTS := ATTEMPTS + 1;
                            OKAY := false
                            end
                        end (* not CH in ['Y','y'] *)
                  until CH in ['Y','y','N','n']
                  end (* EXISTS(FILENAME) *)
                else FileCreate(FILENAME.INPSTR)
              until OKAY;
            Assign(F,FILENAME.INPSTR);
            rewrite(F)  (* output file *)
            end (* O *)
        end (*case*)
  end;  (* GETFILE *)
(* END MODULE GETFILE         VERSION= 'XYLEMMODS     Version  October 7, 2022'; *)

   (**********************************************************)
   (*  WORD I/O Procedures.                                  *)
   (**********************************************************)
(* BEGIN MODULE INPWORD *)
  (* Read a WORD from the terminal. Does NOT read the rest of the line. *)
   procedure INPWORD(var W:WORD);
      var I : integer;
         CH : char;
         BACKSPACE: char;
      begin
        with W do begin
          BACKSPACE:= chr(8);
          LEN:= 0;
          if not(eoln(input)) then read(input,CH);
          while (CH = ' ') do read(input,CH);
          LEN := 1;
          STR[LEN] := CH;
          while (CH <> ' ') and not(eoln(input)) do begin
            read(input,CH);
            if (CH= BACKSPACE) and (LEN > 0) then begin
              write(BACKSPACE);LEN:= LEN-1 end
            else if LEN < MAXWORD then begin
              LEN := LEN + 1;
              STR[LEN]:= CH
            end 
          end;
        if eoln(input) then readln;
        for I := LEN+1 to MAXWORD do STR[I] := ' '
        end
      end; (* INPWORD *)
(* END MODULE INPWORD *)
 

(* BEGIN MODULE READWORD *)
  (* Read a WORD from file F. Does NOT read the rest of the line. 
     If a word can't be read, a 0 length word is returned. *)
   procedure READWORD(var F:text;var W:WORD);
      var I : integer;
         CH, NEWLINE, BLANK : char;

      (* return next character from the file, or newline if eoln *)         
      function NEXTCHAR(var F:text):char;
         var CH : char;
         begin
             if not eoln(F) then begin
                 read(F,CH);
                 NEXTCHAR := CH
             end
             else NEXTCHAR := NEWLINE
         end;
         
      begin
        NEWLINE := chr(10);
        BLANK := chr(32);
        with W do begin
          LEN:= 0;
          CH := NEXTCHAR(F);
          while (CH = BLANK) do CH:= NEXTCHAR(F); (* skip leading blanks *)
          while (CH <> BLANK) and (CH <> NEWLINE) and (LEN < MAXWORD) do begin
              LEN := LEN + 1;
              STR[LEN] := CH;  
              CH := NEXTCHAR(F)        
              end;          
          for I := LEN+1 to MAXWORD do STR[I] := BLANK
        end
      end; (* READWORD *)
(* END MODULE READWORD *)

 
(* BEGIN MODULE WORDPOSN *)
   (*  return the leftmost position of CH in W 
     return -1 if not found. *)
   function WORDPOSN(W:WORD;CH:char):integer;
     var I :integer;
         FOUND: integer;
     begin
       FOUND := -1;
       I := 1;
       with W do begin
         while (I <= LEN) and (FOUND = -1) do begin
           if (STR[I] = CH) then FOUND := I;
           I := I + 1
           end (* while *)
       end; (* with *)   
       WORDPOSN := FOUND       
     end; (* WORDPOSN *)
(* END MODULE WORDPOSN VERSION= 'CSAPMODS     Version  3/17/2023'; *)

(* BEGIN MODULE WORD2LONGINT *)
function WORD2LONGINT(W:WORD; LBOUND,HBOUND:longint):longint;

   var I,VAL,ORDZERO  : integer;
     LEGAL,INRANGE,NEGATIVE      : boolean;
     CH : char;
     NUM : longint;
 
  begin
    INRANGE:= false;
    LEGAL := false;
    NEGATIVE := false;
    ORDZERO:= ord('0');

        with W do begin
 
          (* Evaluate sign, if any *)
          I:=1;CH:=STR[I];
          if CH = '-' then begin
             NEGATIVE := true;
             I:=I+1; CH:= STR[I]
             end
          else begin NEGATIVE:= false; if CH='+' then begin
                         I:=I+1; CH:= STR[I] end end;
 
          (* Evaluate unsigned integer *)
          NUM:= 0;
          while CH in ['0'..'9'] do
            begin
              VAL:= ord(CH)-ORDZERO;
              NUM:= (NUM * 10) + VAL;
              I:= I+1; CH:= STR[I]
            end;
          if I > LEN then LEGAL:= true
          else  begin
              LEGAL:= false;
              writeln;
              writeln('Illegal character encountered.');
              Halt
             end
         end; (* with *)

       (* writeln(NUM); *)
       (* If the number entered was negative, multiply *)
       (* NUM by -1. Check range of number.            *)
       if NEGATIVE then NUM:= - NUM;
       if (NUM >= LBOUND) and (NUM <= HBOUND)
         then INRANGE := true
       else begin
         INRANGE := false;
         Halt
       end;
    WORD2LONGINT := NUM    

   end;   
(* END MODULE WORD2LONGINT *)
 
(* BEGIN MODULE WRITEWORD *)
   (*  Write a word to a file using L characters, left-justified.  *)
   procedure WRITEWORD(var F:text;W:WORD;L:integer);
     var I :integer;
     begin
       with W do
         for I := 1 to L do
           if I <= LEN then write(F,STR[I])
           else write(F,' ')
     end;
(* END MODULE WRITEWORD         VERSION= 'CSAPMODS     Version  4/13/88'; *)
 
(* BEGIN MODULE GETWORD *)
  (*  Read a word from LINE L *)
  procedure GETWORD(CLINE:LINE;var W:WORD;var P:integer);
    var I : integer;
    begin
      with W do begin
        LEN:=0;
        while CLINE.STR[P]=' ' do P:=P+1; 
        while CLINE.STR[P] <> ' ' do begin 
          if LEN < MAXWORD then begin
            LEN := LEN + 1;
            STR[LEN]:= CLINE.STR[P]
            end;
          P:= P+1
          end;
        for I := LEN+1 to MAXWORD do STR[I]:= ' '
        end
    end; (* GETWORD *)
(* END MODULE GETWORD *)

(* BEGIN MODULE TOUPPER *)
    (* Change a character from lower to uppercase *)
      function TOUPPER(CH:char):char;
       begin
       if not(CH in ['a'..'z']) then TOUPPER:=CH
       else case CH of
        'a':TOUPPER:='A'; 'b':TOUPPER:='B'; 'c':TOUPPER:='C'; 'd':TOUPPER:='D';
        'e':TOUPPER:='E'; 'f':TOUPPER:='F'; 'g':TOUPPER:='G'; 'h':TOUPPER:='H';
        'i':TOUPPER:='I'; 'j':TOUPPER:='J'; 'k':TOUPPER:='K'; 'l':TOUPPER:='L';
        'm':TOUPPER:='M'; 'n':TOUPPER:='N'; 'o':TOUPPER:='O'; 'p':TOUPPER:='P';
        'q':TOUPPER:='Q'; 'r':TOUPPER:='R'; 's':TOUPPER:='S'; 't':TOUPPER:='T';
        'u':TOUPPER:='U'; 'v':TOUPPER:='V'; 'w':TOUPPER:='W'; 'x':TOUPPER:='X';
        'y':TOUPPER:='Y'; 'z':TOUPPER:='Z'
       end
      end; (* TOUPPER *)
(* END MODULE TOUPPER *)

(* BEGIN MODULE TOLOWER *)
    (* Change a character from upper to lowercase *)
      function TOLOWER(CH:char):char;
       begin
       if not(CH in ['A'..'Z']) then TOLOWER:=CH
       else case CH of
        'A':TOLOWER:='a'; 'B':TOLOWER:='b'; 'C':TOLOWER:='c'; 'D':TOLOWER:='d';
        'E':TOLOWER:='e'; 'F':TOLOWER:='f'; 'G':TOLOWER:='g'; 'H':TOLOWER:='h';
        'I':TOLOWER:='i'; 'J':TOLOWER:='j'; 'K':TOLOWER:='k'; 'L':TOLOWER:='l';
        'M':TOLOWER:='m'; 'N':TOLOWER:='n'; 'O':TOLOWER:='o'; 'P':TOLOWER:='p';
        'Q':TOLOWER:='q'; 'R':TOLOWER:='r'; 'S':TOLOWER:='s'; 'T':TOLOWER:='t';
        'U':TOLOWER:='u'; 'V':TOLOWER:='v'; 'W':TOLOWER:='w'; 'X':TOLOWER:='x';
        'Y':TOLOWER:='y'; 'Z':TOLOWER:='z'
       end
      end; (* TOLOWER *)
(* END MODULE TOLOWER *)

(* BEGIN MODULE GETCHAR *)
    (* Read a character from the console and check *)
    (*  for correct response.                      *)
    procedure GETCHAR(var CH:char;PNAME:LETTERS;ALLOWED:CHSET);
      const MAXATTEMPTS = 5;
            CTRLC = chr(3); (* ASCII ETX (end fo text, ctrl-C *)
      var   ATTEMPTS : integer;
      begin
        writeln;
        ATTEMPTS := 0;
        repeat
          writeln('Type new value for ',PNAME,'  (CURRENT VALUE: ',CH,')');
          read(CH);
          if CH = CTRLC then halt
          else CH:=TOUPPER(CH);
          readln;
          if not (CH in ALLOWED) then writeln('Inappropriate response: ',CH);
          ATTEMPTS := ATTEMPTS + 1
        until (CH in ALLOWED) or (ATTEMPTS > MAXATTEMPTS)
      end; (* GETCHAR *)
(* END MODULE GETCHAR *)
 
(* BEGIN MODULE GETREAL *)
 (* The procedure GETREAL has been adapted from the procedure 'rdr', from
    Jensen,K., and Wirth,N. (1974) Pascal User Manual and Report, 2nd
    Ed., Springer-Verlag, pp122-123.  The scope of real numbers considered
    legal by GETREAL includes all possible real numbers as defined on p111
    with two additional allowances: (i) numbers may begin with a decimal
    point (ii) 'E' or 'e' may be used to indicate exponentiation. *)
 
    procedure GETREAL(var VAL:real;LOW,HIGH:real);
 
    label 776,777,778;
 
(*!!!*)    const  HIGHEXP = 38;
(*!!!*)           LOWEXP = -38;
                  MAXATTEMPTS = 3;
    type   EXPONENT = 0..HIGHEXP;
 
     var   CH: char;
           LIMIT: real;
           Y : real;
           ORDZERO,I,E, ATTEMPTS : integer;
           LEGAL,INRANGE :boolean;
           NUMWORD  : WORD;
           NEG1,NEG2 : boolean;
           A,S : real;
           LCOUNT,RCOUNT : integer;
           DIGITS: set of '0'..'9';
 
 
      function TEN(E:EXPONENT):real; (* = 10**E *)
        var I:integer;
            T:real;
        begin
          I:=0; T:=1.0;
          repeat
            if odd(E) then
              case I of
                0: T:= T * 1.0E1;
                1: T:= T * 1.0E2;
                2: T:= T * 1.0E4;
                3: T:= T * 1.0E8;
                4: T:= T * 1.0E16;
                5: T:= T * 1.0E32;
        (*!!!   6: T:= T * 1.0E64; Max. exponent is 38
                7: T:= T * 1.0E128; 
                8: T:= T * 1.0E256; *)
                end;
            E:= E div 2; I:= I+1
          until E = 0;
          TEN:= T
        end; (* TEN *)
 
    begin
      LIMIT:= MAXREAL/10;
      DIGITS:=['0'..'9']; ORDZERO:= ord('0');
      ATTEMPTS := 0;
      repeat
       ATTEMPTS := 0;
       repeat
        LEGAL:= true; INRANGE:= true;
        INPWORD(NUMWORD);
        with NUMWORD do begin
          (* Evaluate sign, if any *)
          I:=1;
          CH:= STR[I];
          if CH= '-' then begin NEG1:= true; I:= I+1; CH := STR[I] end
          else begin NEG1:= false; if CH='+' then begin
                                         I:= I+1; CH:= STR[I] end end;
          if not (CH in ['0'..'9','.']) then goto 777;
 
          (* Evaluate whole number part (optional) *)
          A:= 0; E:= 0;
          LCOUNT:= 1;
          while CH in DIGITS do begin
            if A < LIMIT then A:= 10*A + (ord(CH)-ORDZERO)
                          else E:= E+1;
            I:= I+1; CH:= STR[I];LCOUNT:=LCOUNT+1
            end;
 
          (* Evaluate fractional part. *)
          RCOUNT:=0;
          if CH = '.' then begin
            I:= I+1; CH:= STR[I];
            while CH in DIGITS do begin
              if A < LIMIT then
                 begin A:= 10*A + (ord(CH)-ORDZERO); E:= E-1 end;
              I:=I+1; CH:= STR[I]; RCOUNT:= RCOUNT+1
              end
            end;
 
          (* Evaluate exponent *)
          if ((CH='E') or (CH='e')) then begin
            I:= I+1; CH:= STR[I]; S:= 0;
            if CH= '-' then begin NEG2:=true;I:= I+1;CH:=STR[I] end
                       else begin NEG2:=false;
                       if CH='+' then begin I:=I+1;CH:=STR[I] end
                       end;
            if CH in DIGITS then begin
              S:= ord(CH)-ORDZERO;
              I:=I+1; CH:= STR[I];
              while CH in DIGITS do begin
                if S < LIMIT then S:= 10*S + (ord(CH)-ORDZERO);
                I:=I+1; CH:= STR[I]
                end
              end
            else goto 777;
            if NEG2 then E:= E-round(S) else E:= E+round(S)
            end;
 
          (* Check for errors  *)
          (* if I <= LEN then (*illegal char*) goto 777; *)
          if I < LEN then (*illegal char*) goto 777;
          if ((E-RCOUNT<LOWEXP) or (E+LCOUNT>HIGHEXP)) then goto 776;
 
          (* Calculate final value *)
          Y:= A;
          if NEG1 then Y:= -Y;
          if E < 0 then VAL:= Y/(TEN(-E))
           else if E > 0 then VAL:= Y*(TEN(E))
           else VAL:=Y;
          LEGAL:= true;
          goto 778;
 
          (* Error handling statements *)
      776:LEGAL:= false;writeln;writeln(' Exponent Error.');
          writeln(' Enter new value:'); goto 778;
      777:LEGAL:=false;writeln;
          writeln(' Illegal character encountered.');
          write( 'Enter new value:' );
      778:; (* no error, do nothing *)
       end; (* with *)
       ATTEMPTS := ATTEMPTS + 1
     until LEGAL or (ATTEMPTS > MAXATTEMPTS);
     if ATTEMPTS > MAXATTEMPTS then begin
         writeln('Too many attempts. Program terminating');
         Halt
     end;
 
     (* Make sure number is in range *)
     if ((VAL<LOW) or (VAL>HIGH)) then begin
       INRANGE:= false;
       writeln;
       writeln(' Number is out of range');
       writeln(' Enter new value:')
       end
     else INRANGE:= true;
     ATTEMPTS := ATTEMPTS + 1
    until INRANGE or (ATTEMPTS > MAXATTEMPTS);
    if (ATTEMPTS > MAXATTEMPTS) then begin
         writeln('Too many attempts. Program terminating');
         Halt
    end;
   end; (* GETREAL *)
(* END MODULE GETREAL *)
 
(* BEGIN MODULE GETINTEGER *)
  (* Prompts user for an integer, checks whether all characters are digits,*)
  (* and whether number is within desired range; harasses user until valid *)
  (* integer is received. *)
  procedure GETINTEGER(var NUM:longint;LBOUND,HBOUND:longint);
 
   const
     MAXATTEMPTS = 3;
   var I,VAL,ORDZERO,ATTEMPTS  : integer;
     LEGAL,INRANGE,NEGATIVE      : boolean;
     CH : char;
     NUMWORD : WORD;
 
  begin
    INRANGE:= false;
    LEGAL := false;
    NEGATIVE := false;
    ORDZERO:= ord('0');
    ATTEMPTS := 0;
    repeat
      ATTEMPTS := 0;
      repeat
        INPWORD(NUMWORD);
        with NUMWORD do begin
 
          (* Evaluate sign, if any *)
          I:=1;CH:=STR[I];
          if CH = '-' then begin
             NEGATIVE := true;
             I:=I+1; CH:= STR[I]
             end
          else begin NEGATIVE:= false; if CH='+' then begin
                         I:=I+1; CH:= STR[I] end end;
 
          (* Evaluate unsigned integer *)
          NUM:= 0;
          while CH in ['0'..'9'] do
            begin
              VAL:= ord(CH)-ORDZERO;
              NUM:= (NUM * 10) + VAL;
              I:= I+1; CH:= STR[I]
            end;
          if I > LEN then LEGAL:= true
          else  begin
              LEGAL:= false;
              writeln;
              writeln('Illegal character encountered.');
              writeln('Enter new value:  ');
             end
         end; (* with *)
         ATTEMPTS := ATTEMPTS + 1
       until LEGAL or (ATTEMPTS > MAXATTEMPTS);
       if ATTEMPTS > MAXATTEMPTS then begin
           writeln('Too many attempts. Program terminating');
           Halt
           end;
       (* writeln(NUM); *)
       (* If the number entered was negative, multiply *)
       (* NUM by -1. Check range of number.            *)
       if NEGATIVE then NUM:= - NUM;
       if (NUM >= LBOUND) and (NUM <= HBOUND)
         then INRANGE := true
       else begin
         INRANGE := false;
         writeln;
         writeln('Number is out of range.');
         writeln('Please enter new value:')
         end;
       ATTEMPTS := ATTEMPTS + 1
     until INRANGE or (ATTEMPTS > MAXATTEMPTS);
    if (ATTEMPTS > MAXATTEMPTS) then begin
         writeln('Too many attempts. Program terminating');
         Halt
    end
   end;   (* GETINTEGER *)
(* END MODULE GETINTEGER *)
 
(* BEGIN MODULE NUC *)
  (*****************************************************************)
  (*  Convert a character to the appropriate nucleotide symbol.    *)
  (*****************************************************************)
  function NUC(CH:char):NUCLEOTIDE;
    begin
      case CH of
       'A','a': NUC:= A;'C','c': NUC:= C;'G','g': NUC:= G; 
       'T','t','U','u': NUC:= T;'R','r': NUC:= R;'M','m': NUC:= M;
       'B','b': NUC:= B;'N','n': NUC:= N;'Y','y': NUC:= Y;
       'K','k': NUC:= K;'D','d': NUC:= D;'S','s': NUC:= S;
       'W','w': NUC:= W;'H','h': NUC:= H;'V','v': NUC:= V
       end
    end;
(* END MODULE NUC         VERSION= 'CSAPMODS     Version  4/13/88'; *)
 
(* BEGIN MODULE AA *)
  (*****************************************************************)
  (*  Convert a character to the appropriate amino acid symbol.    *)
  (*****************************************************************)
  function AA(CH:char):AMINOACID;
    begin
      case CH of
        'A':AA:=ALA; 'C':AA:=CYS; 'D':AA:=ASP;    'E':AA:=GLU; 'F':AA:=PHE;
        'G':AA:=GLY; 'H':AA:=HIS; 'I':AA:=ILE;    'K':AA:=LYS; 'L':AA:=LEU;
        'M':AA:=MET; 'N':AA:=ASN; 'P':AA:=PRO;    'Q':AA:=GLN; 'R':AA:=ARG;
        'S':AA:=SER; 'T':AA:=THR; 'V':AA:=VALINE; 'W':AA:=TRP; 'Y':AA:=TYR;
        'B':AA:=ASX; 'Z':AA:=GLX; 'X':AA:=UNKX;   '*':AA:=TERM
       end
    end; (* AA *)
(* END MODULE AA *)

(* BEGIN MODULE SAMEWORD *)
      (* Compare two WORDS for equality *)
      function SAMEWORD(var W1,W2:WORD):boolean;
        var I:integer;
            T:boolean;
        begin
          if W1.LEN = W2.LEN then begin
            T:=true;I:=1;
            while (I <= W1.LEN) and T do
              if W1.STR[I] = W2.STR[I] then I:=I+1
              else T:=false;
            SAMEWORD:=T
            end 
          else SAMEWORD:=false
        end; (* SAMEWORD *)
(* END MODULE SAMEWORD  9/12/91 *)

(* BEGIN MODULE READSEQ *)
    (*  ******************************************* *)
    (*  Read a DNA or RNA sequence from SFILE       *)
    (*   and store it in S.                         *)
    (*  ******************************************* *)
    procedure READSEQ(var SFILE:text; var S:SEQUENCE; var SEQLEN:longint;
                      var NAME:WORD; var CIRCULAR:boolean);
      const
          MAXATTEMPTS = 5;
      var CH,FILETYPE,LASTNUM: char;
          TOOLONG,BADFILE:boolean;
          ID,LOCUS,ORI:WORD;
          I,ATTEMPTS:integer;
      begin
        (* Prompt user for sequence file type *)    
        writeln('The following file formats can be read:');
        writeln('  R:free (raw) format   F:FASTA    G:GENBANK');
        ATTEMPTS := 0;
        repeat
          writeln('Type letter of format (R|F|G)');
          readln(CH);
          ATTEMPTS := ATTEMPTS + 1
        until (CH in ['R','r','F','f','G','g']) or (ATTEMPTS > MAXATTEMPTS);
        if (ATTEMPTS > MAXATTEMPTS) then begin
             writeln('Too many attempts. Program terminating');
             Halt
        end;
        case CH of
          'R','r':FILETYPE:='R';
          'F','f':FILETYPE:='F';
          'G','g':FILETYPE:='G'
          end; (* case *)
        writeln('Reading input file...');
        BADFILE:=false;
        NAME.LEN:=0;
        
        (* For FASTA or GENBANK, read in sequence name and topology *)
        (* Advance to beginning of sequence *)
        if FILETYPE= 'F' then begin
          (* First non-comment line is name. Name may be blank.*)
          read(SFILE,CH);
          if CH = '>' then begin
              READWORD(SFILE,NAME); readln(SFILE)
              end
          end (* FASTA *)
          
        else if FILETYPE='G' then begin
          (* Initialize identifiers *)
          with LOCUS do begin
            STR[1]:='L';STR[2]:='O';STR[3]:='C';STR[4]:='U';STR[5]:='S';
            LEN:=5
            end;
          with ORI do begin
            STR[1]:='O';STR[2]:='R';STR[3]:='I';STR[4]:='G';STR[5]:='I';
            STR[6]:='N';LEN:=6
            end;
          
           (* Advance to LOCUS line. Read in NAME and topology. *)
           (* while not((SFILE^='L') or (eof(SFILE))) do readln(SFILE);*)
           if not eof(SFILE) then begin
             READWORD(SFILE,ID);
             (*WRITEWORD(output,ID,ID.LEN); write(' '); write(ID.LEN); writeln;*)
             if SAMEWORD(ID,LOCUS) then begin
                (*writeln('Reading locus name...');*)
                READWORD(SFILE,NAME); 
                (*WRITEWORD(output,NAME,NAME.LEN); write(' ');*)
                (* skip past sequence length, units, and seq. type *)
                READWORD(SFILE,ID);
                READWORD(SFILE,ID);
                READWORD(SFILE,ID);
                (* read topology *)
                READWORD(SFILE,ID);
                CH:=ID.STR[1];
                if (CH='C') or (CH='c') then CIRCULAR:=true
                else CIRCULAR:=false;
                (*write(CH); writeln;*)
                end (* SAMEWORD(ID,LOCUS) *)
              else BADFILE:=true
             end;
          
           (* Advance to ORIGIN line. Sequence begins on next line *)
           if not eof(SFILE) then begin
             (*writeln('Looking for ORIGIN line...');*)
             repeat
               readln(SFILE);
               READWORD(SFILE,ID)
             until SAMEWORD(ID,ORI) or eof(SFILE);
             (*if SAMEWORD(ID,ORI) then WRITEWORD(output,ID,ID.LEN);*)
             if SAMEWORD(ID,ORI) then readln(SFILE)
             end;
           if eof(SFILE) then BADFILE:= true
           end; (* GENBANK *)
        
        (* Read in sequence *)
        SEQLEN := 0;TOOLONG:=false;
        if not BADFILE then begin
          LASTNUM := '0';
          while not eof(SFILE) do begin
            while not eoln(SFILE) do begin
              read(SFILE,CH);
              if CH in ['A','a','C','c','G','g','T','t','U','u','N','n',
                        'R','r','Y','y','M','m','W','w','S','s','K','k',
                        'D','d','H','h','V','v','B','b'] then
                if SEQLEN < MAXSEQ-2 then begin
                (*write(CH);*)
                  SEQLEN := SEQLEN + 1;
                  S[SEQLEN]:= NUC(CH);
                  end
                else TOOLONG:= true
              else if CH=';' then (*begin    comment in input file *)
                                    readln(SFILE)(*;writeln end*)
              else if CH in ['1','2'] then LASTNUM:=CH                      
              end;
            readln(SFILE)(*writeln*)
            end; (* if not BADFILE *)
          if TOOLONG then writeln(
            '>>> WARNING! Sequence length exceeds MAXSEQ-2. Seq. truncated.');

          if FILETYPE in ['R','F'] then begin
              if LASTNUM in ['1','2'] then begin
                  case LASTNUM of
                      '1':CIRCULAR:= false;
                      '2':CIRCULAR:= true
                      end (* case CH *)
                  end
              else begin 
                  ATTEMPTS := 0;
                  repeat
                      writeln('Is sequence circular or linear? (Type C or L)');
                      readln(CH);
                      ATTEMPTS := ATTEMPTS + 1
                  until (CH in ['C','c','L','l']) or (ATTEMPTS > MAXATTEMPTS);
                  if (ATTEMPTS > MAXATTEMPTS) then begin
                      writeln('Too many attempts. Program terminating');
                      Halt
                      end;
                  case CH of
                      'C','c':CIRCULAR:= true;
                      'L','l':CIRCULAR:= false
                      end (* case CH *)
                  end (* if LASTNUM in ['1','2'] *)
               end
           end (* if FILETYPE in ['R','F'] *)
         else writeln('>>> ERROR: Not a GENBANK file. No sequence read.');

      Close(SFILE)
      end; (* READSEQ *)
(* END MODULE READSEQ *)

(* BEGIN MODULE READPRO *)
    (*  ******************************************* *)
    (*  Read a protein    sequence from SFILE       *)
    (*   and store it in PR.                        *)
    (*  ******************************************* *)
    procedure READPRO(var SFILE:text; var PR:PROTEIN; var LEN:longint;
                      var NAME:WORD);

      label 86;

      const
          MAXATTEMPTS = 5; (* Max. no. of attempts at terminal input before halting *)

      var CH,ANSWER,FILETYPE: char;
          J,ATTEMPTS: integer;
          TOOLONG:boolean;
 
      begin
        (* Prompt user for sequence file type *)
        writeln('The following file formats can be read:');
        writeln('  R:free (raw) format    F:Fasta');
        ATTEMPTS := 0;
        repeat
          writeln('Type letter of format (F|N)');
          readln(ANSWER);
          ATTEMPTS := ATTEMPTS + 1
        until (ANSWER in ['R','r','F','f']) or (ATTEMPTS > MAXATTEMPTS);
        (* prevent infinite loop *)
        if (ATTEMPTS > MAXATTEMPTS) then begin
             writeln('Too many attempts. Program terminating');
             Halt
        end;
        case ANSWER of
          'R','r':FILETYPE:='R';
          'F','f':FILETYPE:='F'
          end;
        writeln('Reading input file...');

        NAME.LEN:=0;

        (* Read name from the first line of a FASTA file *)
        if FILETYPE = 'F' then begin
          if not eoln(SFILE) then read(SFILE,CH);
          if CH = '>' then begin
            READWORD(SFILE,NAME)
            end; (* with NAME *)
            readln(SFILE)
          end;

        (* Read in the sequence *)
        J := 0;TOOLONG:=false;
        while not eof(SFILE) do begin
          while not eoln(SFILE) do begin
            read(SFILE,CH); CH := TOUPPER(CH);
            if CH in ['G','A','V','L','I','M','F','P','S','T','C','N','Q',
                      'Y','W','D','E','H','K','R','B','Z','*','X'] then
              if J < MAXSEQ-MAXRANGE then begin
                J := J + 1;
                case CH of
                  'G':PR[J]:= GLY; 'A':PR[J]:= ALA; 'V':PR[J]:= VALINE;
                  'L':PR[J]:= LEU; 'I':PR[J]:= ILE; 'M':PR[J]:= MET;
                  'F':PR[J]:= PHE; 'P':PR[J]:= PRO; 'S':PR[J]:= SER;
                  'T':PR[J]:= THR; 'C':PR[J]:= CYS; 'N':PR[J]:= ASN;
                  'Q':PR[J]:= GLN; 'Y':PR[J]:= TYR; 'W':PR[J]:= TRP;
                  'D':PR[J]:= ASP; 'E':PR[J]:= GLU; 'H':PR[J]:= HIS;
                  'K':PR[J]:= LYS; 'R':PR[J]:= ARG; 'X':PR[J]:= UNKX; 
                  'B':PR[J]:= ASX; 'Z':PR[J]:= GLX; 
                  '*':begin  
                        PR[J]:= TERM;
                        if FILETYPE='N' then goto 86 (*ignore rest of file*)
                      end
                  end
                end
              else TOOLONG:= true
            else if CH=';' then readln(SFILE) (* comment in input file *)
            end;
          readln(SFILE)
          end;

(* !!! *) 86: Close(SFILE);
        if TOOLONG then writeln(
          '>>> WARNING! Sequence exceeds MAXSEQ-MAXRANGE. Sequence truncated.');
        LEN := J
      end; (* READPRO *)
(* END MODULE READPRO *) 

(* BEGIN MODULE LINKED *)
  (*********************************************************)
  (*  Linked-list operations for restriction fragment list.*)
  (*********************************************************)
 
  (*Get a new fragment from freelist.*)
  procedure GETFRAG(var NEWFRAG:FRAG);
    begin
      if FREEFRAG = nil then new(NEWFRAG)
      else begin
        NEWFRAG:= FREEFRAG;
        FREEFRAG:= FREEFRAG^.NEXT
      end
    end;
 
  (*Add a fragment after DEST*)
  procedure ADDFRAG(var AFRAG,DEST:FRAG);
    var TEMP: FRAG;
    begin
      TEMP:= DEST^.NEXT;
      DEST^.NEXT:= AFRAG; AFRAG^.PREV:= DEST;
      AFRAG^.NEXT:= TEMP;TEMP^.PREV:= AFRAG
    end;
 
  (*Return a list to the top of freelist*)
  procedure RIDOF(var HEAD,TAIL:FRAG);
    var TEMPHEAD,TEMPTAIL:FRAG;
    begin
      if HEAD^.NEXT <> TAIL then begin
        TEMPHEAD:= HEAD^.NEXT;TEMPTAIL:= TAIL^.PREV;
        HEAD^.NEXT:= TAIL;TAIL^.PREV:= HEAD;
        TEMPHEAD^.PREV:= nil;TEMPTAIL^.NEXT:= FREEFRAG;
        FREEFRAG:= TEMPHEAD
        end
    end;
(* END MODULE LINKED *)
 
(* BEGIN MODULE SORT *)
  (*  Invariant:  The array elements > TOP are sorted.*)
  (*    TOP >= unsorted elements.                     *)
  procedure BUBBLESORT(TOP,BOTTOM:integer);
  var SMALLEST,NEXT :integer;
  procedure SWAP(var FIRST,SECOND:FRAG);
    var TEMP:FRAG;
    begin
      TEMP:= FIRST;FIRST:=SECOND;SECOND:=TEMP
    end;
  begin
    while TOP >= BOTTOM do begin
      (*bubble smallest unsorted number to the top of sorted list*)
      SMALLEST:= BOTTOM; NEXT:= BOTTOM+1;
      while NEXT <= TOP do begin
        if ORDER[SMALLEST]^.SIZE < ORDER[NEXT]^.SIZE then
          SWAP(ORDER[SMALLEST],ORDER[NEXT]);
          SMALLEST:= NEXT;
          NEXT:= SMALLEST+1
        end;
      TOP:= TOP - 1
      end;
  end; (* BUBBLESORT *)
(* END MODULE SORT *)

(* BEGIN MODULE MV *)
   (*****************************************************************)
   (*  Calculate the mean and variance of first N elements          *)
   (*  in array X.  The type VECTOR must be declared in a higher    *)
   (*  scope as   array[1..<positive constant>] of <real|integer>   *)
   (*****************************************************************)
   procedure MV(var X:VECTOR; N:integer; var MEAN,VARIANCE:real);
     var SX,SSX:real;
         I:integer;
     begin
       SX:=0; SSX:=0;
       for I:= 1 to N do begin
         SX:= SX + X[I];
         SSX:= SSX + sqr(X[I])
         end;
       MEAN:= SX/N;
       VARIANCE:= (SSX - (sqr(SX)/N))/N-1
     end; (* MV *)
(* END MODULE MV *)
  
  (* ----------------------------------------------------------  *)
  (* ----------------- MAIN  PROCEDURE  -----------------------  *)
    begin
(* BEGIN MODULE STARTUP *)
(* Peform operations which must be done at beginning of main
   procedure. *)
(*!!!   TERMIN(input);    Open input for interactive use *)
(*!!!   TERMOUT(output);   "   output "      "        "  *)
     writeln(VERSION:50);
     writeln;
(* END MODULE STARTUP *)
 
      writeln('This program tests the module procedures in CSAPMOD.');
      writeln;
      writeln('TEST OF INTERACTIVE INPUT');
      CH:= ' ';
      writeln('When prompted, type a letter between A and Z');
      GETCHAR(CH,'CHARACTER:',['A'..'Z']);
      writeln('The character is: ',CH);
      writeln('Type a single word and press RETURN');
      INPWORD(WO); 
      write('The word you typed was: ');
      WRITEWORD(output,WO,WO.LEN);writeln;
      writeln('Test of WORDPOSN');
      writeln('Type a character to search for in the word.');
      GETCHAR(CH,'CHARACTER:',['A'..'Z']); 
      RESULT := WORDPOSN(WO,CH);
      writeln(RESULT);         
      writeln('Type a line of text and press RETURN');
      INPLINE(LI);
      writeln('The line you typed was:');
      WRITELINE(output,LI,LI.LEN);
      writeln('Number of characters: ',LI.LEN);
      NUMWORDS := 0; I := 0;
      while (I < LI.LEN) do begin
          NEXTWORD(LI,WO,I); WRITEWORD(output,WO,WO.LEN); write(' ');
          if WO.LEN > 0 then NUMWORDS := NUMWORDS + 1
          end;
      writeln;
      write('Number of words in the line you typed was: ');
      writeln(NUMWORDS);     
      writeln;
      writeln('TEST OF NUMERICAL INPUT');
      writeln('Type an integer between ',MININT,' and ',MAXINT,
              'and press RETURN');
      GETINTEGER(INT,MININT,MAXINT);
      writeln('The integer you typed was: ',INT);
      writeln('Type a real number between ',-MAXREAL,' and ',MAXREAL);
      writeln('and press RETURN');
      GETREAL(RE,-MAXREAL,MAXREAL);
      writeln('The real number you typed was: ',RE);
      writeln;
      writeln('TEST OF FILE I/O');
      writeln('Enter Nucleotide sequence filename:');
      GETFILE(INFILE,'I',LI);
      CIRCULAR := true;
      READSEQ(INFILE,SEQ,SEQLEN,WO,CIRCULAR);
      writeln('The sequence in');
      WRITELINE(output,LI,LI.LEN); write(output, ' named ');WRITEWORD(output,WO,WO.LEN); writeln;
      write('is ',SEQLEN,' bases long.  ');
      if CIRCULAR then writeln('Topology: circular')
      else writeln('Topology: linear');
      writeln('Enter Protien sequence filename:');
      GETFILE(INFILE,'I',LI);
      READPRO(INFILE,PROT,SEQLEN,WO);
      writeln('The sequence in');
      WRITELINE(output,LI,LI.LEN); write(output, ' named ');WRITEWORD(output,WO,WO.LEN); writeln;
      writeln('is ',SEQLEN,' amino acids long.');
      writeln('Test of file creation');
      writeln('Enter name of file to create:');
      GETFILE(OUTFILE,'O',LI);
      writeln(OUTFILE,'Test succeeded.');
      close(OUTFILE)

    end. (* CSAPMOD *)
 
