  (* ********************************************************  *)
  (*                                                           *)
  (*   XYLEM_SHUFFLE   VERSION   8 May 2023  Free Pascal       *)
  (*             Brian Fristensky                              *)
  (*             Dept. of Plant Science                        *)
  (*             University of Manitoba                        *)
  (*             Winnipeg, MB  R3T 2N2 CANADA                  *)
  (*                                                           *)
  (* SYNOPSIS                                                  *)
  (*    xylem_shuffle [-w<n> -o<n>]  <input >output                  *)
  (*                                                           *)
  (* DESCRIPTION                                               *)
  (*    Shuffles nucleic acid or protein sequences in a        *)
  (*    FASTA/Pearson (.wrp) format file. see Lipman et al     *)
  (*                                                           *)
  (*    -s<n>  (DEPRECATED) This option has no effect, but is  *)
  (*           retained for backward compatability.            *)
  (*           The Free Pascal Randomize function initializes  *)
  (*           the random number stream using the system clock.*)
  (*           n is a random integer between 0 and 32767.      *)
  (*           This number must be provided for each run.      *)
  (*                                                           *)
  (*    -w<n>  n is an integer, indicating the width of the    *)
  (*           window for random localization. It should       *)
  (*           never exceed the length of the shortest input   *)
  (*           sequence.                                       *)
  (*                                                           *)
  (*    -o<n>  n is an integer, indicating the number of nuc-  *)
  (*           leotides overlap between adjacent windows. It   *)
  (*           should never exceed the window size.  o def-    *)
  (*           aults to 0 of not specified.                    *)
  (*                                                           *)
  (*    If w and o are specified, overlapping windows of w     *)
  (*    nucleotides are shuffled, thus preserving the local    *)
  (*    characteristic base composition. Windows overlap by    *)
  (*    o nucleotides.                                         *)
  (*                                                           *)
  (*    If w and o are not specified, each sequence is shuf-   *)
  (*    fled globally, thus preserving the overall base        *)
  (*    composition, but not the local variations in comp.     *)
  (*                                                           *)
  (* Copyright (c) 1988, 2023  by Brian Fristensky.            *)
  (* !!! in comment indicates feature which may need change.   *)
  (*  *******************************************************  *)
 (* Modification history:
   8 May 2023  o Adapted code from Standard Pascal to Free Pascal
               o Name changed to XYLEM_SHUFFLE to avoid potential
                 name conflicts.
               o The option for setting a random seed has been
                 removed from the command line. The FPC Randomize
                 function does that using the system clock.
               o The previous version printed program parameters
                 as a separate comment line. To better comply with
                 FASTA format, these parameters are simply added
                 to the name line.
               o To distinguish the randomized sequence from the
                 original sequence, "-rand" is appended to the 
                 sequence name in the output.
*) 

(*!!!*)program XYLEM_SHUFFLE(input, output);
 (*!!! Some Pascals require file parameters in heading *)

uses 
   custApp;
 
  const MAXSEQ  =  750000;
        MAXWORD = 25;
        MAXLINE = 120;
        STARTARGNUM=1;
(* BEGIN MODULE REALLIMITS *)
(*!!!*)MAXREAL = 1.7E38;
(* END MODULE REALLIMITS *)

        VERSION = 'XYLEM_SHUFFLE      Version 8 May 2023';
 
  type SEQCHAR  = 
                (T,U,C,A,G,N,R,Y,M,W,S,K,D,H,V,B,L,Z,F,P,E,I,Q,ASTERISK,X,GAP);
       SEQUENCE    = array[1..MAXSEQ] of SEQCHAR;
      
(* BEGIN MODULE TYPE.WORD *)
       (*   <word>::= <non-blank char>[<non-blank char>] *)
       WORD    = record
                 LEN:integer;
                 STR:array[1..MAXWORD] of char
                 end;
(* END MODULE TYPE.WORD *)

(* BEGIN MODULE TYPE.LINE *)
       CHARARRAY = array[1..MAXLINE] of char;
       LINE = record
                INPSTR:string;
                STR:CHARARRAY;
                LEN:integer
                end;
(* END MODULE TYPE.LINE *)
 
  var SEQ          : SEQUENCE;
      SEQNAME      : WORD;
      SEQLEN       : integer;
      SCHAR       : array[SEQCHAR] of char;
      (* global processing parameters *)
      ARGNUM,
      POS:integer;
      CH : char;
      UNREADOPTIONS:boolean;
      ARGUMENT:CHARARRAY; 
      RSEED,OVERLAP,WINDOW: integer;
      J:integer;

(* BEGIN MODULE NUMBER *)
  (* Extract an integer from a CHARARRAY (see TYPE.LINE), starting
     at the current position. *)
  function NUMBER(TARGET:CHARARRAY; var POS:integer):longint;
    var N:longint;
        ORDZERO:integer;
    begin
      ORDZERO:=ord('0');
      N:=0;
      (* evaluate characteristic *)
      while TARGET[POS] in ['0'..'9'] do begin
        N:= (N*10) + (ord(TARGET[POS])-ORDZERO);
        POS:=POS+1
        end; (* while *)
      NUMBER:=N
    end; (* NUMBER *)
  function NUMBER(TARGET:CHARARRAY; var POS:integer):integer;
    var N:integer;
        ORDZERO:integer;
    begin
      ORDZERO:=ord('0');
      N:=0;
      (* evaluate characteristic *)
      while TARGET[POS] in ['0'..'9'] do begin
        N:= (N*10) + (ord(TARGET[POS])-ORDZERO);
        POS:=POS+1
        end; (* while *)
      NUMBER:=N
    end; (* NUMBER *)
(* END MODULE NUMBER *)
 
(* BEGIN MODULE READWORD *)
  (* Read a WORD from file F. Does NOT read the rest of the line. 
     If a word can't be read, a 0 length word is returned. *)
   procedure READWORD(var F:text;var W:WORD);
      var I : integer;
         CH, NEWLINE, BLANK : char;

      (* return next character from the file, or newline if eoln *)         
      function NEXTCHAR(var F:text):char;
         var CH : char;
         begin
             if not eoln(F) then begin
                 read(F,CH);
                 NEXTCHAR := CH
             end
             else NEXTCHAR := NEWLINE
         end;
         
      begin
        NEWLINE := chr(10);
        BLANK := chr(32);
        with W do begin
          LEN:= 0;
          CH := NEXTCHAR(F);
          while (CH = BLANK) do CH:= NEXTCHAR(F); (* skip leading blanks *)
          while (CH <> BLANK) and (CH <> NEWLINE) and (LEN < MAXWORD) do begin
              LEN := LEN + 1;
              STR[LEN] := CH;  
              CH := NEXTCHAR(F)        
              end;          
          for I := LEN+1 to MAXWORD do STR[I] := BLANK
        end
      end; (* READWORD *)
(* END MODULE READWORD *)

(* BEGIN MODULE WRITEWORD *)
   (*  Write a word to a file using L characters, left-justified.  *)
   procedure WRITEWORD(var F:text;W:WORD;L:integer);
     var I :integer;
     begin
       with W do
         for I := 1 to L do
           if I <= LEN then write(F,STR[I])
           else write(F,' ')
     end;
(* END MODULE WRITEWORD *)
           
(* BEGIN MODULE READLINE *)
  (* Read a line from a file, omitting trailing blanks *)  
  procedure READLINE(var F:text; var L:LINE);
    var LASTNONBLANK,I:integer;
        CH:char;
    begin
      with L do begin
        LEN:=0; LASTNONBLANK:=0;
        while not eoln(F) do begin
          read(F,CH);
          if LEN < MAXLINE then begin 
            LEN:=LEN+1; STR[LEN]:=CH;
            if CH <> ' ' then LASTNONBLANK:=LEN
            end
          end;
        if not eof(F) then readln(F);
        LEN:=LASTNONBLANK;
        for I:= LEN+1 to MAXLINE do STR[I]:=' '
        end; (* with L*) 
    end; (* READLINE *)
(* END MODULE READLINE *)

(* BEGIN MODULE WRITELINE *)
   (*  Write a line to a file using L char, left-justified.  *)
   procedure WRITELINE(var F:text;W:LINE;L:integer);
     var I :integer;
     begin
       with W do
         for I := 1 to L do
           if I <= LEN then write(F,STR[I])
           else write(F,' ')
     end; (* WRITELINE *)
(* END MODULE WRITELINE *)
 
  (*****************************************************************)
  (*  Convert a character to the appropriate sequence symbol.      *)
  (*****************************************************************)
  function NUCPROT(CH:char):SEQCHAR;
    begin
      case CH of
       'A','a': NUCPROT:= A;'C','c': NUCPROT:= C;'G','g': NUCPROT:= G; 
       'U','u': NUCPROT:=U; 'L','l':NUCPROT:=L;
       'T','t': NUCPROT:= T;'R','r': NUCPROT:= R;'M','m': NUCPROT:= M;
       'B','b': NUCPROT:= B;'N','n': NUCPROT:= N;'Y','y': NUCPROT:= Y;
       'K','k': NUCPROT:= K;'D','d': NUCPROT:= D;'S','s': NUCPROT:= S;
       'W','w': NUCPROT:= W;'H','h': NUCPROT:= H;'V','v': NUCPROT:= V;
       'Z','z':NUCPROT:=Z;'F','f':NUCPROT:=F;'P','p':NUCPROT:=P;
       'E','e':NUCPROT:=E;'I','i':NUCPROT:=I;'Q','q':NUCPROT:=Q;
       '*':NUCPROT:=ASTERISK;'X','x':NUCPROT:=X;'-':NUCPROT:=GAP;
       end
    end;

  (*****************************************************)
  (* Initialization procedures.                        *)
  (*****************************************************)
  procedure INITIALIZE;
    begin
      (*Set default values for parameters *)
      WINDOW:=0; OVERLAP:=0; RSEED:=0;
 
      (*   Initialize SCHAR array which holds the          *)
      (*   character values of the SEQCHARS.               *)
      SCHAR[A]:='A'; SCHAR[C]:='C';SCHAR[G]:='G'; SCHAR[T]:='T';
      SCHAR[U]:='U'; SCHAR[L]:='L';
      SCHAR[R]:='R'; SCHAR[Y]:='Y';SCHAR[M]:='M'; SCHAR[W]:='W';
      SCHAR[S]:='S'; SCHAR[K]:='K';SCHAR[D]:='D'; SCHAR[H]:='H';
      SCHAR[V]:='V'; SCHAR[B]:='B';SCHAR[N]:='N'; SCHAR[Z]:='Z';
      SCHAR[F]:='F'; SCHAR[P]:='P';SCHAR[E]:='E'; SCHAR[I]:='I';
      SCHAR[Q]:='Q'; SCHAR[X]:='X';SCHAR[ASTERISK]:='*'; SCHAR[GAP]:='-';
 
     end; (* INITIALIZE *)


    (* - - - - - - - - - - - - - - - - - - - - - - - - - - -*)
    (* Read in the next sequence                            *)
     procedure READSEQUENCE(var SEQFILE:text; var CH:char; 
                   var SEQNAME:WORD; var SEQ:SEQUENCE; var SEQLEN:integer);
        var TOOLONG:boolean;
        begin
            if CH = '>' then begin READWORD(SEQFILE,SEQNAME);readln(SEQFILE) end;
            read(SEQFILE,CH);
            SEQLEN := 0; TOOLONG:=false;
            while (not eof(SEQFILE)) and (not (CH = '>')) do begin
                if CH in ['A','a','C','c','G','g','T','t','U','u','N','n',
                        'R','r','Y','y','M','m','W','w','S','s','K','k',
                        'L','l',
                        'D','d','H','h','V','v','B','b','Z','z','F','f',
                        'P','p','E','e','I','i','Q','q','*','X','x','-'] then
                    if SEQLEN < MAXSEQ-2 then begin
                        SEQLEN := SEQLEN + 1;
                        SEQ[SEQLEN]:=NUCPROT(CH);
                        end
                    else TOOLONG:= true;
                if eoln(SEQFILE) then readln(SEQFILE);
                read(SEQFILE,CH)
                end; (* eof *)
              if TOOLONG then writeln(
             '>>> WARNING! Sequence length exceeds MAXSEQ-2. Seq. truncated.')
        end; (* READSEQUENCE *)

  (* ******************************************************************* *)
  (*  Randomize each sequence by shuffling groups of WINDOW bases .      *)
  (*  Each adjecent window overlaps by OVERLAP positions.                *)
  (* ******************************************************************* *)
   procedure SCRAMBLE(var SEQ:SEQUENCE; SEQLEN:integer; WINDOW:integer);
 
     var LEFT,RIGHT,OFFSET:integer;

     (* Randomize nucleotides/aa's within a window by swapping. *)
     procedure SCRAMBLE_WINDOW(LEFT,RIGHT:integer);

       var I,J:integer;
           RNUM:real;

       (* Swap nucleotides or amino acids between two positions. *)
       procedure SWAP(var N1,N2:SEQCHAR);
         var TEMP:SEQCHAR;
         begin
           TEMP:=N1;
           N1:=N2;
           N2:=TEMP
         end; (* SWAP *)

       begin (* SCRAMBLE_WINDOW *)
         for I:= LEFT to RIGHT do begin
           RNUM:= Random();
           J:=LEFT + round(RNUM * WINDOW);
           if J>RIGHT then J:=RIGHT; (* Make sure to stay within bounds *)
           SWAP(SEQ[I],SEQ[J])
           end
       end; (* SCRAMBLE_WINDOW *)

     begin (* SCRAMBLE *)
       LEFT:=1;
       if WINDOW < 1 then WINDOW:=SEQLEN
       else if WINDOW > SEQLEN then WINDOW:=SEQLEN;
       RIGHT:=WINDOW;
       OFFSET:=WINDOW-OVERLAP;

       while RIGHT <= SEQLEN do begin 
         SCRAMBLE_WINDOW(LEFT,RIGHT);
         LEFT:= LEFT+OFFSET;
         RIGHT:=RIGHT+OFFSET;
         if RIGHT > SEQLEN then begin (* end of sequence, a special case *)
           RIGHT:=SEQLEN;
           WINDOW:=RIGHT-LEFT+1;
           if WINDOW>1 then SCRAMBLE_WINDOW(LEFT,RIGHT);
           RIGHT:=RIGHT+OFFSET;
           end (* RIGHT > SEQLEN *)
         end (* RIGHT <= SEQLEN *)
     end; (* SCRAMBLE *)
 
   (* - - - - - - - - - - - - - - - - - - - - - - - - - - -*)
   procedure WRITESEQ(var OUTFILE:text; SEQ:SEQUENCE; SEQNAME:WORD; SEQLEN:integer);
     const LINELEN = 50;
     var SEQPRINTED,THISLINE:integer;
     begin
       (* write nameline, with -rand appended to name, and shuffling paramters *)
       write(output,'>');
       WRITEWORD(output,SEQNAME,SEQNAME.LEN);
       write(OUTFILE,'-rand    ');
       writeln(output,'XYLEM_SHUFFLE:   WINDOW=',WINDOW,' ':5,'OVERLAP=',OVERLAP);

       (* write sequence *) 
       SEQPRINTED:=0;
       while SEQPRINTED < SEQLEN do begin
         THISLINE:=SEQPRINTED + LINELEN;
         if THISLINE > SEQLEN then THISLINE:=SEQLEN;
         while SEQPRINTED < THISLINE do begin
           SEQPRINTED:=SEQPRINTED + 1;
           write(OUTFILE,SCHAR[SEQ[SEQPRINTED]])
           end;
         writeln(OUTFILE)
         end (* SEQPRINTED < SEQLEN *)
      end; (* WRITESEQ *)     

  (* ----------------------------------------------------------  *)
  (* ----------------- MAIN  PROCEDURE  -----------------------  *)
    begin
      INITIALIZE;

      (* Read options from the command line. Note that -s is required *)
      ARGNUM:= STARTARGNUM;
      UNREADOPTIONS:=true;
      while UNREADOPTIONS do
        if ARGNUM < argc then begin
          ARGUMENT := paramStr(ARGNUM);
          if ARGUMENT[1]='-' then begin
            if ARGUMENT[2] in ['s','w','o'] then begin
              POS:=3;
              case ARGUMENT[2] of
                (* 's':RSEED:=NUMBER(ARGUMENT,POS); *)
                'w':WINDOW:=NUMBER(ARGUMENT,POS);
                'o':OVERLAP:=NUMBER(ARGUMENT,POS)
                end (* case*)
              end (* s,w,o *)
            end (* '-' *)
          else UNREADOPTIONS:=false;
          ARGNUM:=ARGNUM+1
          end (* if ARGNUM <= argc *)
        else UNREADOPTIONS:=false;
      
      (* Initialize random number stream. *)
      (*RANDSEED(RSEED);*)
      Randomize();
      read(input,CH);
      while not eof(input) do begin
          READSEQUENCE(input,CH,SEQNAME,SEQ,SEQLEN);          
          SCRAMBLE(SEQ,SEQLEN,WINDOW);
          WRITESEQ(output,SEQ,SEQNAME,SEQLEN) 
          end (* not eof(SEQFILE) *)       

    end. (* XYLEM_SHUFFLE *)
